@@ -164,26 +164,59 @@ def list_schemas(cur, pattern, verbose):
         return [(None, cur, headers, cur.statusmessage)]
 
 
+# https://github.com/postgres/postgres/blob/master/src/bin/psql/describe.c#L5471-L5638
 @special_command('\\dx', '\\dx[+] [pattern]', 'List extensions.')
 def list_extensions(cur, pattern, verbose):
 
-    # Note: psql \dx command seems to ignore schema patterns
-    _, name_pattern = sql_name_pattern(pattern)
+    def _find_extensions(cur, pattern):
+        sql = '''
+            SELECT e.extname, e.oid FROM pg_catalog.pg_extension e
+        '''
+
+        params = []
+        if pattern:
+            _, schema = sql_name_pattern(pattern)
+            sql += 'WHERE e.extname ~ %s'
+            params.append(schema)
+
+        sql = cur.mogrify(sql + 'ORDER BY 1, 2;', params)
+        log.debug(sql)
+        cur.execute(sql)
+        return cur.fetchall()
+
+    def _describe_extension(cur, oid):
+        sql = '''
+            SELECT  pg_catalog.pg_describe_object(classid, objid, 0)
+                    AS "Object Description"
+            FROM    pg_catalog.pg_depend
+            WHERE   refclassid = 'pg_catalog.pg_extension'::pg_catalog.regclass
+                    AND refobjid = %s
+                    AND deptype = 'e'
+            ORDER BY 1'''
+        sql = cur.mogrify(sql, [oid])
+        log.debug(sql)
+        cur.execute(sql)
+
+        headers = [x[0] for x in cur.description]
+        return cur, headers, cur.statusmessage
+
+    if cur.connection.server_version < 90100:
+        not_supported = "Server versions below 9.1 do not support extensions."
+        cur, headers = [], []
+        yield None, cur, None, not_supported
+        return
 
     if verbose:
-        extensions = _find_extensions(cur, name_pattern)
+        extensions = _find_extensions(cur, pattern)
 
-        if not extensions:
-            msg = 'Did not find any extension named "%s"' % pattern
-            return [(None, cur, [], msg)]
-
-        results = []
-        for ext_name, oid in extensions:
-            title = 'Objects in extension "%s"' % ext_name
-            cur, headers, status = _describe_extension(cur, oid)
-            results.append((title, cur, headers, status))
-
-        return results
+        if extensions:
+            for ext_name, oid in extensions:
+                title = '\nObjects in extension "%s"' % ext_name
+                cur, headers, status = _describe_extension(cur, oid)
+                yield title, cur, headers, status
+        else:
+            yield None, None, None, 'Did not find any extension named "{}".'.format(pattern)
+        return
 
     sql = '''
       SELECT e.extname AS "Name",
@@ -198,46 +231,18 @@ def list_extensions(cur, pattern, verbose):
                 AND c.classoid = 'pg_catalog.pg_extension'::pg_catalog.regclass
       '''
 
-    if name_pattern:
-        sql = cur.mogrify(
-            sql + ' WHERE e.extname ~ %s ORDER BY 1', [name_pattern])
-    else:
-        sql += ' ORDER BY 1'
-
-    log.debug(sql)
-    cur.execute(sql)
-    headers = [x[0] for x in cur.description]
-    return [(None, cur, headers, cur.statusmessage)]
-
-
-def _find_extensions(cur, pattern):
-    sql = 'SELECT e.extname, e.oid FROM pg_catalog.pg_extension e'
-
+    params = []
     if pattern:
-        sql = cur.mogrify(sql + ' WHERE e.extname ~ %s', [pattern])
-
-    sql += ' ORDER BY 1'
+        _, schema = sql_name_pattern(pattern)
+        sql += 'WHERE e.extname ~ %s'
+        params.append(schema)
 
+    sql = cur.mogrify(sql + 'ORDER BY 1, 2', params)
     log.debug(sql)
     cur.execute(sql)
-    return cur.fetchall()
-
-
-def _describe_extension(cur, oid):
-    sql = '''
-        SELECT  pg_catalog.pg_describe_object(classid, objid, 0)
-                  AS "Object Description"
-        FROM    pg_catalog.pg_depend
-        WHERE   refclassid = 'pg_catalog.pg_extension'::pg_catalog.regclass
-                AND refobjid = %s
-                AND deptype = 'e'
-        ORDER BY 1'''
-    sql = cur.mogrify(sql, [oid])
-    log.debug(sql)
-    cur.execute(sql)
-
-    headers = [x[0] for x in cur.description]
-    return cur, headers, cur.statusmessage
+    if cur.description:
+        headers = [x[0] for x in cur.description]
+        yield None, cur, headers, cur.statusmessage
 
 
 def list_objects(cur, pattern, verbose, relkinds):
@@ -683,15 +688,14 @@ def describe_table_details(cur, pattern, verbose):
     where = []
     params = []
 
-    if not pattern:
-        where.append('pg_catalog.pg_table_is_visible(c.oid)')
+    where.append('pg_catalog.pg_table_is_visible(c.oid)')
 
     if schema:
         where.append('n.nspname ~ %s')
         params.append(schema)
 
     if relname:
-        where.append('c.relname ~ %s')
+        where.append('c.relname OPERATOR(pg_catalog.~) %s')
         params.append(relname)
 
     sql = """SELECT c.oid, n.nspname, c.relname
@@ -1633,3 +1637,44 @@ def shell_command(cur, pattern, verbose):
     cur, headers = [], []
     params = shlex.split(pattern)
     return [(None, cur, headers, subprocess.call(params))]
+
+@special_command('\\dE', '\\dE[+] [pattern]', 'List foreign tables.', aliases=() )
+def list_foreign_tables(cur, pattern, verbose):
+
+    if verbose:
+        verbose_cols = '''
+            , pg_catalog.pg_size_pretty(pg_catalog.pg_table_size(c.oid)) as "Size",
+            pg_catalog.obj_description(c.oid, 'pg_class') as "Description" '''
+    else:
+        verbose_cols = ''
+
+    if pattern:
+        _, tbl_name = sql_name_pattern(pattern)
+        filter = " AND c.relname OPERATOR(pg_catalog.~) '^(%s)$' " % tbl_name
+    else:
+        filter = ''
+
+    query = '''
+        SELECT n.nspname as "Schema",
+        c.relname as "Name",
+        CASE c.relkind WHEN 'r' THEN 'table' WHEN 'v' THEN 'view' WHEN 'm' THEN 'materialized view' WHEN 'i' THEN 'index' WHEN 'S' THEN 'sequence' WHEN 's' THEN 'special' WHEN 'f' THEN 'foreign table' WHEN 'p' THEN 'table' WHEN 'I' THEN 'index' END as "Type",
+        pg_catalog.pg_get_userbyid(c.relowner) as "Owner"
+        %s
+        FROM pg_catalog.pg_class c
+            LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace
+        WHERE c.relkind IN ('f','')
+            AND n.nspname <> 'pg_catalog'
+            AND n.nspname <> 'information_schema'
+            AND n.nspname !~ '^pg_toast'
+        AND pg_catalog.pg_table_is_visible(c.oid)
+        %s
+        ORDER BY 1,2;
+        ''' % (verbose_cols, filter)
+
+    cur.execute(query)
+    if cur.description:
+        headers = [x[0] for x in cur.description]
+        return [(None, cur, headers, cur.statusmessage)]
+    else:
+        return [(None, None, None, cur.statusmessage)]
+
